//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// June 20, 2000
//
#include "IppvImplementation.h"
#include "IppvTranslator.h"

#include <Tracer.h>
#include <errstr.h>
#include <SasDalUtils.h>
#include <DataFilter.h>
#include <cstdio>
#include <dal_utilities.h>

#include <string>
#include <map>
#include <typeinfo>

//
// IppvServer
//

IppvServer::IppvServer() { PROC ENDPROC }

IppvServer::~IppvServer() { PROC ENDPROC }

Ippv * IppvServer::ippv(const string& name,
			const string& colname,
			const Dal::DataType datatype,
			const string& label,
			const Ippv::Type type,
			const bool& singlevalue) const
{
  PROC
  Ippv * r = 0;
  switch(type) {
  case Ippv::Periodic:
    switch(datatype) {
    case Dal::Int8:
      r = new PeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int16:
      r = new PeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint16:
      r = new PeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int32:
      r = new PeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint32:
      r = new PeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real32:
      r = new PeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real64:
      r = new PeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, singlevalue);
      break;
    case Dal::DString:
      r = new PeriodicIppv<string, char, &Data::stringData>(name, colname, label, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;

  case Ippv::NonPeriodic:
    switch(datatype) {
    case Dal::Int8:
      r = new NonPeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int16:
      r = new NonPeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint16:
      r = new NonPeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int32:
      r = new NonPeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint32:
      r = new NonPeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real32:
      r = new NonPeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real64:
      r = new NonPeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, singlevalue);
      break;
    case Dal::DString:
      r = new NonPeriodicIppv<string, char, &Data::stringData>(name, colname, label, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  default:
    Sas::error("UnsupportedHousekeepingParameterType");
    break;
  }
  RETURN r;
}

Ippv * IppvServer::ippv(const string& name,
			const string& colname,
			const Dal::DataType datatype,
			const string& label,
			const Ippv::Type type,
			const string& expression,
			const bool& singlevalue) const
{
  PROC
  Ippv * r = 0;
  switch(type) {
  case Ippv::Periodic:
    switch(datatype) {
    case Dal::Int8:
      r = new PeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int16:
      r = new PeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint16:
      r = new PeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int32:
      r = new PeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint32:
      r = new PeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real32:
      r = new PeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real64:
      r = new PeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::DString:
      r = new PeriodicIppv<string, char , &Data::stringData>(name, colname, label, expression, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  case Ippv::NonPeriodic:
    switch(datatype) {
    case Dal::Int8:
      r = new NonPeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int16:
      r = new NonPeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint16:
      r = new NonPeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int32:
      r = new NonPeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint32:
      r = new NonPeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real32:
      r = new NonPeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real64:
      r = new NonPeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::DString:
      r = new NonPeriodicIppv<string, char , &Data::stringData>(name, colname, label, expression, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  default:
    Sas::error("UnsupportedHousekeepingParameterType");
    break;
  }
  RETURN r;
}

//
// IppvCommon
//
template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::IppvCommon(const string& name,
				   const string& colname,
				   const string& label,
				   const bool& singlevalue)
  : _name(name),
    _colname(colname),
    _expression(""),
    _label(label),
    _filter(false),
    _singlevalue(singlevalue),
    _has_translator(false),
    _trans(new IppvTranslator<T>)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::IppvCommon(const string& name,
				   const string& colname,
				   const string& label,
				   const string& expression,
				   const bool& singlevalue)
  : _name(name),
    _colname(colname),
    _expression(expression),
    _label(label),
    _filter(true),
    _singlevalue(singlevalue),
    _has_translator(false),
    _trans(new IppvTranslator<T>)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::~IppvCommon() { PROC ENDPROC }

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::fetch(const TimeOrderedTable *tab,
				   const STime& start,
				   const STime& stop) throw(IppvException)
{
  PROC

  unsigned rowl = lower_row(tab, start);
  if(rowl == tab->rows()){
    errstr << tab->table().qualifiedName() + " Time range: "
	   << tab->time_of_row(0) << " - "
	   << tab->time_of_row(tab->rows() - 1)
	   << endl << "Time sought: " << start
	   << Sas::message(Msg::AppLib, Msg::Noisy);
    strstream s;
    s << "TimeOutOfRange: lower time bound for " + _name + " in " +
      tab->table().qualifiedName() +
      " not available. Looking for an entry at or around "
      << start << " but the available time range in "
      <<  tab->table().name() << " is "
      << tab->time_of_row(0) << " - "
      << tab->time_of_row(tab->rows() - 1) << ends;
    throw OutOfRange(s.str());
  }
  unsigned rowu = upper_row(tab, stop);
  if(rowu == tab->rows()){
    strstream s;
    s << "TimeOutOfRange: upper time bound for " + _name + " in " +
      tab->table().qualifiedName() +
      " not available. Looking for an entry at or around "
      << stop << " but the available time range in "
      <<  tab->table().name() << " is "
      << tab->time_of_row(0) << " - "
      << tab->time_of_row(tab->rows() - 1) << ends;
    throw OutOfRange(s.str());
  }
  
  vector<DataFilter::RowRange> rr;
  if(_filter){
    strstream s;
    // Perverse!! #ROW start from 1 but the DAL starts from 0
    s << "(" + _expression + ") && (#ROW >= " << rowl+1
      << " && #ROW <= " << rowu+1 << ")" << ends;
    string expression = s.str();
    DataFilter df(expression);
    df.doSelect(tab->table(), rr);
    if(rr.size() == 0)
      throw NoHkRecords("NoAvailableRows: No rows match the expression "
			+ expression + " in " + tab->table().qualifiedName());
  } else {
    DataFilter::RowRange r;
    r.startRow = rowl;
    r.stopRow = rowu;
    rr.push_back(r);
  }

  errstr << "start/stop " << start << " / " << stop << Sas::message(Msg::AppLib, Msg::Noisy);
  for(unsigned k=0; k<rr.size(); ++k){
    errstr << "Row range = [" << k << "] " << rr[k].startRow << " - " << rr[k].stopRow
	   << Sas::message(Msg::AppLib, Msg::Noisy);
  }

  //  _label += " [" + tab->table().column(_colname)->units() + "]";
  try {
//      if(_singlevalue && rr.size() > 1){
    // setValue below does not know anything about _singlevalue. So we
    // make sure that the row range is mododified here.
    if(_singlevalue){
      rr.erase(rr.begin() + 1, rr.end());
      rr[0].stopRow = rr[0].startRow;
      Sas::message(Msg::AppLib, Msg::Noisy, "RowRange modified.");
    }
    setValue(tab->table().column(_colname), rr);
  }
  catch (IppvException e) {
    throw;
  }
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
unsigned IppvCommon<T, S, data>::lower_row(const TimeOrderedTable *,
					   const STime&) const
{
  PROC
  Sas::error("NotImplementedHere", "this call is not implemented here");
  RETURN 0;
}

template<class T, class S, S * (Data::*data)() const>
unsigned IppvCommon<T, S, data>::upper_row(const TimeOrderedTable *,
				       const STime&) const
{
  PROC
  Sas::error("NotImplementedHere", "this call is not implemented here");
  RETURN 0;
}

template<class T> static string dalTypeToString(T v)
{
  static bool firstTime = true;
  static map<string, string> m;
  if(firstTime){
    m[typeid(bool).name()] = toString(Dal::Bool);
    m[typeid(int8).name()] = toString(Dal::Int8);
    m[typeid(uint16).name()] = toString(Dal::Uint16);
    m[typeid(int16).name()] = toString(Dal::Int16);
    m[typeid(uint32).name()] = toString(Dal::Uint32);
    m[typeid(int32).name()] = toString(Dal::Int32);
    m[typeid(real32).name()] = toString(Dal::Real32);
    m[typeid(real64).name()] = toString(Dal::Real64);
    m[typeid(dstring).name()] = toString(Dal::DString);
    m[typeid(string).name()] = "string";
    firstTime = false;
  }
  if(m.find(typeid(v).name()) != m.end())
    return m[typeid(v).name()];
  else
    return typeid(v).name();
}
// FIXME: duplicated code in the specialization

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::print(ostream& os, const bool verbose) const
{
  PROC
  if(verbose){
    os << _name << " " << dalTypeToString<T>(_value) << " [" << _colname ;
    if(_filter)
      os << " subject to " << _expression;
    os << "]";
    if(_has_translator){
      os << endl << "This parameter is translated as follows:" << endl;
      _trans->print(os);
    }
  } else {
    os << _name << " = ";
    if(_has_translator){
      try{
	os << _trans->look_up(_value);
      }
      catch(NoTranslationAvailable e){
	e.print(errstr);
	errstr << "It is not possible to translate " + _name
	  + ". Its raw value is " << _value
	       << Sas::warning("TranslationNotAvailable");
      }
    }
    else
      os << _value;
    os << " // " << _label;
  }
  os << endl;
  ENDPROC
}


// Specialization required because in the DAL int8 = char and << will
// try to print a character instead of an unsigned integer
template<>
void IppvCommon<int8, int8, &Data::int8Data>::print(ostream& os, const bool verbose) const
{
  PROC
  if(verbose){
    os << _name << " " << dalTypeToString<int8>(_value) << " [" << _colname ;
    if(_filter)
      os << " subject to " << _expression;
    os << "]";
    if(_has_translator){
      os << "This parameter is translated as follows:" << endl;
      _trans->print(os);
    }
  } else {
    os << _name << " = ";
    if(_has_translator){
      try{
	os << _trans->look_up(_value);
      }
      catch(NoTranslationAvailable e){
	e.print(errstr);
	errstr << "It is not possible to translate " + _name
	  + ". Its raw value is " << _value
	       << Sas::warning("TranslationNotAvailable");
	throw;
      }
    }
    else
      os << static_cast<int>(_value);
    os << " // " << _label;
  }
  os << endl;
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::notAvailable(ostream& os) const
{
  PROC
  os << "// WARNING: " << _name << " not found" << endl;
  os << _name + "= Unknown // " + _label << endl;
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
const string& IppvCommon<T, S, data>::name() const
{
  PROC
    RETURN _name;
}

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::setValue(const Column *column,
				      const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  // ATTENTION: column->data() returns ColumnData*, but the pointer to
  // member stuff only works for the base class Data. I do not know if
  // this is a bug in gcc of it is supposed to be this way.
  Data* cd = column->data();
  _value = (cd->*data)()[r[0].startRow];
  for(unsigned i = 0; i < r.size(); ++i)
    for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
      T v;
      if((v = (cd->*data)()[k]) != _value){
	strstream s;
	s << "Parameter " + _name + " from column " + _colname
	  << ". Expected " << _value << " but found " << v
	  << " at row " << k
	  << ends;
	delete cd;
	throw ParameterChange(s.str());
      }
    }
  delete cd;
  ENDPROC
}

// Specialization for real-valued cases, when we calculate the mean of
// the values in the range. 

template<>
void IppvCommon<real32, real32, &Data::real32Data>::setValue(const Column *column,
							     const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  // ATTENTION: column->data() returns ColumnData*, but the pointer to
  // member stuff only works for the base class Data. I do not know if
  // this is a bug in g++ of it is supposed to be this way.
  Data* cd = column->data();
  real32 v = 0.;
  unsigned nrows = 0;
  for(unsigned i = 0; i < r.size(); ++i)
    for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
      ++nrows;
      v += (cd->*&Data::real32Data)()[k];
    }
  _value = v/nrows;
  delete cd;
  ENDPROC
}

// Specialization required because reading rows from a character
// column requires a different mechanism.
template<>
void IppvCommon<string, char, &Data::stringData>::setValue(const Column *column,
							   const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  ColumnData* cd = column->data();
  _value = SasDalUtils::getStringColumnRow(cd, r[0].startRow);
  for(unsigned i = 0; i < r.size(); ++i)
    for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
      string v;
      if((v = SasDalUtils::getStringColumnRow(cd, k)) !=
	   _value){
	  strstream s;
	  s << "Parameter " + _name + " from column " + _colname
	    << ". Expected " << _value << " but found " << v
	    << " at row " << k
	    << ends;
	  delete cd;
	  throw ParameterChange(s.str());
	}
    }
  delete cd;
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T,S,data>::addTranslator(IppvTranslator<T> *t)
{
  PROC
  _trans = t;
  _has_translator = true;
  ENDPROC
}


//
// PeriodicIppv
//
template<class T, class S, S * (Data::*data)() const>
PeriodicIppv<T, S, data>::PeriodicIppv(const string& name,
				       const string& colname,
				       const string& label,
				       const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, singlevalue)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
PeriodicIppv<T, S, data>::PeriodicIppv(const string& name,
				       const string& colname,
				       const string& label,
				       const string& expression,
				       const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, expression, singlevalue)
{
  PROC
  ENDPROC
}

//  template<class T, class S, S * (Data::*data)() const>
//  PeriodicIppv<T, S, data>::~PeriodicIppv() { PROC ENDPROC }
 
template<class T, class S, S * (Data::*data)() const>
unsigned PeriodicIppv<T, S, data>::lower_row(const TimeOrderedTable *tab,
					     const STime& t) const
{
  PROC
  RETURN tab->lower_bound(t);
}

template<class T, class S, S * (Data::*data)() const>
unsigned PeriodicIppv<T, S, data>::upper_row(const TimeOrderedTable *tab,
					     const STime& t) const
{
  PROC
  RETURN tab->upper_bound(t);
}

//
// NonPeriodicIppv
//
template<class T, class S, S * (Data::*data)() const>
NonPeriodicIppv<T, S, data>::NonPeriodicIppv(const string& name,
					     const string& colname,
					     const string& label,
					     const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, singlevalue)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
NonPeriodicIppv<T, S, data>::NonPeriodicIppv(const string& name,
					     const string& colname,
					     const string& label,
					     const string& expression,
					     const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, expression, singlevalue)
{
  PROC
  ENDPROC
}

//  template<class T, class S, S * (Data::*data)() const>
//  NonPeriodicIppv<T, S, data>::~NonPeriodicIppv() { PROC ENDPROC }
 
template<class T, class S, S * (Data::*data)() const>
unsigned NonPeriodicIppv<T, S, data>::lower_row(const TimeOrderedTable *tab,
						const STime& t) const
{
  PROC
  RETURN tab->first_less_than(t);
}

template<class T, class S, S * (Data::*data)() const>
unsigned NonPeriodicIppv<T, S, data>::upper_row(const TimeOrderedTable *tab,
						const STime& t) const
{
  PROC
  RETURN tab->last_less_than(t);
}

//

template void IppvCommon<uint16, uint16, &Data::uint16Data>::addTranslator(IppvTranslator<uint16> *);
template void IppvCommon<int16, int16, &Data::int16Data>::addTranslator(IppvTranslator<int16> *);
template void IppvCommon<string, char, &Data::stringData>::addTranslator(IppvTranslator<string> *);
template void IppvCommon<real32, real32, &Data::real32Data>::addTranslator(IppvTranslator<real32> *);

// These are needed on DEC/OSF
template class PeriodicIppv<uint16, uint16, &Data::uint16Data>;
template class PeriodicIppv<string, char,   &Data::stringData>;
template class PeriodicIppv<int16,  int16,  &Data::int16Data>;
template class PeriodicIppv<real32, real32, &Data::real32Data>;
